from typing import Union

from visualizr.anitalker.diffusion.diffusion import (
    SpacedDiffusionBeatGans,
    SpacedDiffusionBeatGansConfig,
)

Sampler = Union[SpacedDiffusionBeatGans]
SamplerConfig = Union[SpacedDiffusionBeatGansConfig]
